package beasts;

import java.awt.*;

/**
 * Basic beasts.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: BasicBeast.java,v 1.14 2002/02/09 23:26:50 peteg Exp $
 */
class BasicBeast extends Beast {
    static Image image0, image1;

    /**
     * Claustrophobia threshold. If a beast cannot move for this many turns,
     * it explodes.
     *
     * @see BasicBeast. claustrophobia_count
     */
    private static final int CLAUSTROPHOBIA_THRESHOLD = 5;

    /**
     * Claustrophobia count.
     *
     * @see BasicBeast.CLAUSTROPHOBIA_THRESHOLD
     */
    int claustrophobia_count;

    BasicBeast(BeastsWorld beasts, int x, int y) {
	super(beasts, x, y);
	claustrophobia_count = 0;
    }

    Image getImage() {
	return (claustrophobia_count < CLAUSTROPHOBIA_THRESHOLD / 2)
	    ? image0
	    : image1;
    }

    int points() {
	return (claustrophobia_count > CLAUSTROPHOBIA_THRESHOLD)
	    ? 3
	    : 2;
    }

    public final void move() {
	if(dead) {
	    /* Damn, we got squashed. No more turns for us. */
	    return;
	}

	int[] dirs = new int[2 * 8];
	int dirs_count;

	dirs_count = beasts.getValidDirections(x, y, dirs);
	if(dirs_count > 0) {
	    /* The beast can move, if it desires to. */

	    /* Compute the vector towards the player. */
	    int px = beasts.player.x - x,
		py = beasts.player.y - y;

	    claustrophobia_count = 0;
	    doMove(dirs_count, dirs, px, py);
	} else {
// 	    System.err.println("Beast: dirs_count == 0");

	    // Can't move, so bump the claustrophobia count.
	    if(++claustrophobia_count > CLAUSTROPHOBIA_THRESHOLD) {
		beasts.explode(this);
		beasts.update();
		return;
	    } else {
		// Redraw: the beast may have changed colour.
		beasts.draw(x, y, getImage());
	    }
	}

	beasts.update();
	beasts.add(this, beasts.BEAST_DELAY);
    }

    /**
     * It's possible for this beast to move (i.e. dirs_count > 0). This
     * method implements the tactics of the beasts.
     */
    void doMove(int dirs_count, int[] dirs, int px, int py) {
	/* Simplistic: move towards the player. */
	/* Squash the vector to {-1, 0, 1} ^ 2 */
	px = (px == 0) ? 0 : (px < 0) ? -1 : 1;
	py = (py == 0) ? 0 : (py < 0) ? -1 : 1;

	if(!vetMove(px, py)) {
	    /* Can't move towards the player, so choose a random valid
	     * direction. */
	    int dir = 2 * beasts.rnd(dirs_count);

// 	    System.err.println("Beast: random: dirs_count: "
// 			       + dirs_count + " dir: " + dir / 2);

	    beasts.move(this, dirs[dir], dirs[dir + 1]);
	}
    }

    /**
     * Check that the move is legal, and if so, perform it.
     *
     * @arg dx The change in the <i>x</i> direction (-1..1)
     * @arg dy The change in the <i>y</i> direction (-1..1)
     */
    final boolean vetMove(int dx, int dy) {
// 	System.err.println("vM: dx: " + dx + " dy: " + dy);
	WorldObject wo = beasts.getWO(x + dx, y + dy);

	if(dx == 0 && dy == 0) {
	    return true;
	} else if(wo == null) {
	    beasts.move(this, x + dx, y + dy);
	    return true;
	} else if(wo instanceof Player) {
	    beasts.player.kill();
	    beasts.move(this, x + dx, y + dy);
	    killed_player();
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * We just killed the player.
     */
    void killed_player() {
    }

    public String toString() {
	return "BasicBeast[" + x + ", " + y + "]";
    }
}
