package beasts;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 * AWT context for the graphics, and game thread: start the game, progress
 * levels, move the beasts around the board, schedule egg hatchings, etc.
 * We need this as we can't attach a <code>KeyListener</code> to an
 * <code>Applet</code> subclass to get keyboard events (empirically, anyway,
 * under IBM JDK118).
 * <p>
 * Agents (i.e. the <code>Player</code> and <code>Beast</code> subclasses)
 * grab the lock on an instance of this class to ensure mutex.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: BeastsWorld.java,v 1.15 2002/04/10 07:56:01 peteg Exp $
 */
public class BeastsWorld extends Canvas implements Runnable {
    /* Image dimensions. */
    final static int I_WIDTH = 16;
    final static int I_HEIGHT = 16;

    /* Character dimensions. */
    final static int C_WIDTH = 8;
    final static int C_HEIGHT = I_HEIGHT;

    /* Dimensions of the playing area. */
    final static int ROWS = 23;
    final static int COLS = 40;

    final static int WIDTH  = COLS * I_WIDTH;
    final static int HEIGHT = (ROWS + 2) * BeastsWorld.I_HEIGHT;

    final static int QUIT_GAME = 1;
    final static int QUIT_LEVEL = 2;

    private int quit = 0;

    /* Default to level 'E'. */
    private final static int INIT_LEVEL = 'E' - 'A';

    private final static int STATUSLINE = 23;

    private final static int HEAVY_BLOCKS = 8;
    private final static int LIGHT_BLOCKS = (int)(BeastsWorld.ROWS * BeastsWorld.COLS * 0.25);

    /* Initial delay times, in milliseconds. */
    private final static int INIT_BEAST_DELAY = 1000;
    private final static int INIT_EGG_DELAY = 10000;
    private final static int INIT_EGG_HATCH_DELAY = 100;
    private final static int INIT_SPEEDUP_DELAY = 30000;

    private final static int MAX_EGG_HATCHING_SPEED = 100;

    /* Delay times, in milliseconds. Note that speedup() fiddles with
     * these. */
    int BEAST_DELAY;
    int EGG_DELAY;
    int EGG_HATCH_DELAY;
    int SPEEDUP_DELAY;

    /* Multiplicative factor of speedup, i.e. all delays are multiplied by
     * this number everytime a <code>Speedup</code> event occurs. */
    private static float speedup_factor = 0.9f;

    /* Track the number of beasts so we know when to exit. */
    private int count;
    private int level;
    private int round;
    private int winningRound;

    private PriorityQueue pq;

    Player player;
    boolean training;

    private PersistentState ps;
    private Scores scores;

    private WorldObject[][] worldObjects;

    private Thread thisThread;

    /* Off-screen buffer stuff. */
    private Dimension dim;
    private Graphics offGraphics;
    private Image offImage, offImage2;

    private Image splash;
    private Image ascii;

    private Random random;
    private MediaTracker tracker;

    /**
     * Maps int directions -> vectors. This is an enumeration of
     * co-ordinates, clockwise, from the upper-left, which is labelled
     * (-1,-1).
     */
    private static int[][] dirs = {
	{-1, -1}, {0, -1}, {1, -1}, {1, 0}, {1, 1}, {0, 1}, {-1, 1}, {-1, 0},
	{0, 0}
    };

    /**
     * Maps vectors (+1) -> int directions.
     */
    private static int[][] vecs = { {0, 7, 6}, {1, 8, 5}, {2, 3, 4} };

    /**
     * Convert directions to numbers.
     */
    static int dir2int(int x, int y) {
// 	System.err.println("dir2int: direction: ("
// 			   + x + ", " + y + ") -> "
// 			   + vecs[x + 1][y + 1]);
	return vecs[x + 1][y + 1];
    }

    /**
     * Convert numbers to directions.
     */
    static int[] int2dir(int dir) {
// 	System.err.println("int2dir: int: " + dir + " -> ("
// 			   + dirs[dir][0] + ", " + dirs[dir][1] + ")");
	return dirs[dir];
    }

    /**
     * Squash a vector, then convert it to a number.
     */
    static int squashdir2int(int x, int y) {
	/* Squash the vector to {-1, 0, 1} ^ 2 */
	x = (x == 0) ? 0 : (x < 0) ? -1 : 1;
	y = (y == 0) ? 0 : (y < 0) ? -1 : 1;

	return dir2int(x, y);
    }

    static int abs(int x) {
	return x > 0 ? x : -x;
    }

    public Insets getInsets() {
	return new Insets(0, 0, 0, 0);
    }

    public Dimension getMinimumSize() {
        return dim;
    }

    public Dimension getPreferredSize() {
        return dim;
    }

    public boolean isFocusTraversable() {
	return true;
    }

    public BeastsWorld(PersistentState ps, boolean training) {
	this.dim = new Dimension(WIDTH, HEIGHT);
	this.ps = ps;
	this.training = training;
	this.scores = ps.getScores();
	this.level = INIT_LEVEL;

	/* Scheduler stuff. */
	pq = new PriorityQueue();

	/* Create the random number generator. */
	random = new Random();

	/* Load the graphics. */
        tracker = new MediaTracker(this);

  	offImage = splash = getImage("BeastsSplash.gif");
  	ascii = getImage("VGAfont.gif");
 	BasicBeast.image0 = getImage("BasicBeast0.gif");
 	BasicBeast.image1 = getImage("BasicBeast1.gif");
 	Egg.image0 = getImage("Egg0.gif");
 	Egg.image1 = getImage("Egg1.gif");
 	HatchedSuperBeast.image = getImage("HatchedSuperBeast.gif");
 	LightBlock.image = getImage("LightBlock.gif");
 	Player.image = getImage("Player.gif");
 	HeavyBlock.image = getImage("HeavyBlock.gif");
 	SuperBeast.image = getImage("SuperBeast.gif");

	this.player = new Player(this);

// 	System.err.println("init(): " + Thread.currentThread().toString());
    }

    /**
     * Wrapper function to handle bugs in Netscape 4.x
     * (i.e. <code>getResource()</code> is not implemented).
     */
    private Image getImage(String name) {
	Image image;

//   	image = Toolkit.getDefaultToolkit().getImage(Beasts.class.getResource(name));
        try { 
	    InputStream in;

	    if((in = getClass().getResourceAsStream(name)) == null) {
		System.err.println("Image not found: " + name); 
		throw new RuntimeException("BeastsWorld.getImage(" + name + ") not found.");
	    }

	    byte[] buffer = new byte[in.available()]; 
	    in.read(buffer); 
	    image = Toolkit.getDefaultToolkit().createImage(buffer); 
	} catch (IOException e) { 
	    System.err.println("Unable to read image: " + name); 
	    e.printStackTrace(); 
	    return null;
	} 

	tracker.addImage(image, 0);
	return image;
    }

    /**
     * Set the level to play and display it's characteristics.
     */
    void setLevel(int level) {
	this.level = level;

	offImage = offImage2;
	Levels.displayLevel(this, level);
	update();
    }

    /**
     * Setup the world for a particular level.
     *
     * @see beasts.Levels
     */
    void initLevel(boolean pullBlocks,
		   int winningRound) {
	scores.reset();
	player.initLevel(pullBlocks);
	this.winningRound = winningRound;
    }

    /**
     * Play a level with the specified characteristics.
     *
     * @see beasts.Levels
     */
    void initRound(boolean gameSpeedUp,
		   boolean explosiveBlocks,
		   int superBeastRatio,
		   int eggs,
		   int hatchingSpeed,
		   int startingBeasts) {
	int START_DELAY;
	int SUPER_BEASTS = (superBeastRatio > 0)
	    ? startingBeasts / superBeastRatio
	    : 0;
	startingBeasts -= SUPER_BEASTS;

	this.count = 0;

	if(training) {
	    BEAST_DELAY = 0;
	} else {
	    BEAST_DELAY = INIT_BEAST_DELAY;
	}
	EGG_DELAY = (INIT_EGG_DELAY * hatchingSpeed) / MAX_EGG_HATCHING_SPEED;
	EGG_HATCH_DELAY = INIT_EGG_HATCH_DELAY;
	SPEEDUP_DELAY = INIT_SPEEDUP_DELAY;

	offImage = offImage2;
	clear();

	/* Setup the world. */
	worldObjects = new WorldObject[COLS][ROWS];

	/* Add a border of immovable blocks. */
	for(int i = 0; i < COLS; i++) {
	    worldObjects[i][0]        = new HeavyBlock(this, i, 0,        explosiveBlocks);
	    worldObjects[i][ROWS - 1] = new HeavyBlock(this, i, ROWS - 1, explosiveBlocks);
	}

	for(int i = 1; i < ROWS - 1; i++) {
	    worldObjects[0][i]        = new HeavyBlock(this, 0,        i, explosiveBlocks);
	    worldObjects[COLS - 1][i] = new HeavyBlock(this, COLS - 1, i, explosiveBlocks);
	}

	int[] xy = new int[2];
	int x, y;

	/* Toss in a certain number of immovable blocks in random positions. */
	for(int i = 0; i < HEAVY_BLOCKS; i++) {
	    rndPos(xy); x = xy[0]; y = xy[1];
	    worldObjects[x][y] = new HeavyBlock(this, x, y, explosiveBlocks);
	}

	/* Toss in a certain number of movable blocks in random positions. */
	for(int i = 0; i < LIGHT_BLOCKS; i++) {
	    rndPos(xy); x = xy[0]; y = xy[1];
	    worldObjects[x][y] = new LightBlock(this, x, y);
	}

	/* Add the beasts in random positions. */
	if(startingBeasts > 0) {
	    START_DELAY = BEAST_DELAY / startingBeasts;

	    for(int i = 0; i < startingBeasts; i++) {
		rndPos(xy); x = xy[0]; y = xy[1];
		Beast beast = ps.newBasicBeast(this, x, y);
		worldObjects[x][y] = beast;
		add(beast, BEAST_DELAY + i * START_DELAY);
	    }
	}

	if(SUPER_BEASTS > 0) {
	    START_DELAY = BEAST_DELAY / SUPER_BEASTS;

	    for(int i = 0; i < SUPER_BEASTS; i++) {
		rndPos(xy); x = xy[0]; y = xy[1];
		SuperBeast superbeast = new SuperBeast(this, x, y);
		worldObjects[x][y] = superbeast;
		add(superbeast, BEAST_DELAY + i * START_DELAY);
	    }
	}

	/* Toss in a certain number of eggs in random positions. */
	for(int i = 0; i < eggs; i++) {
	    rndPos(xy); x = xy[0]; y = xy[1];
	    Egg egg = new Egg(this, x, y);
	    worldObjects[x][y] = egg;
	    add(egg, EGG_DELAY + rnd(EGG_DELAY / 2));
	}

	/* Schedule a game speedup, if needs be. */
	if(gameSpeedUp) {
	    pq.enqueue(new Speedup(this), SPEEDUP_DELAY);
	}

	update();
    }

    /**
     * Generate a random, unoccupied location in the world.
     */
    private void rndPos(int[] xy) {
	int x, y;

	do {
	    x = rnd(COLS);
	    y = rnd(ROWS);
	} while(worldObjects[x][y] != null);

	xy[0] = x;
	xy[1] = y;
    }

    /**
     * Enqueue a <code>Beast</code> object, and call it's
     * <code>move()</code> method after the specified <code>delay</code>.
     */
    void add(Beast beast, int delay) {
	count++;
	pq.enqueue(beast, delay);
    }

    void addPlayer() {
	int[] xy = new int[2];
	int x, y;

	/* Ensure the player ends up on a different square than where it
	 * currently is. */
	do {
	    rndPos(xy); x = xy[0]; y = xy[1];
	} while(player.x == x && player.y == y);

	player.x = x;
	player.y = y;
	worldObjects[x][y] = this.player;
	draw(x, y, player.getImage());
    }

    /**
     * Clear the entire offGraphics object.
     */
    void clear() {
	offGraphics.setColor(Color.black);
	offGraphics.fillRect(0, 0, dim.width, dim.height);
    }

    /**
     * Clear a particular square.
     */
    void clear(int x, int y) {
	offGraphics.setColor(Color.black);
	offGraphics.fillRect(x * I_WIDTH, y * I_HEIGHT,
			     I_WIDTH, I_HEIGHT);
    }

    /**
     * Draw a specified image in a particular square.
     */
    void draw(int x, int y, Image image) {
	offGraphics.drawImage(image,
			      x * I_WIDTH, y * I_HEIGHT,
// 			      I_WIDTH, I_HEIGHT,
			      null);
    }

    /**
     * Work out which squares can be moved to from a particular square. The
     * array passed as an argument is filled out with valid absoute
     * locations as pairs: <code>(a[0], a[1]), (a[2], a[3]), ...</code>
     *
     * @return the number of valid directions.
     */
    int getValidDirections(int x, int y, int[] dirs) {
	int xp, yp;
	int count = 0;

	for(int i = 0; i < 8; i++) {
	    int[] xy = int2dir(i);
	    xp = xy[0]; yp = xy[1];

// 	    System.err.println("gVD: xp = " + xp + " yp = " + yp);

	    if(worldObjects[x + xp][y + yp] == null) {
		dirs[2 * count] = x + xp;
		dirs[2 * count + 1] = y + yp;
		count++;
	    }
	}

	return count;
    }

    /**
     * Move an object, update it's position fields to reflect this.
     */
    void move(WorldObject wo, int x, int y) {
	if(worldObjects[x][y] != null) {
	    throw new RuntimeException("Beasts.move: trying to move: " + wo
				       + " to square occupied by: " + worldObjects[x][y]);
	}

	clear(wo.x, wo.y);
	draw(x, y, wo.getImage());

	worldObjects[wo.x][wo.y] = null;
	worldObjects[x][y] = wo;
	wo.x = x;
	wo.y = y;
    }

    /**
     * Look at a square in the world.
     */
    WorldObject getWO(int x, int y) {
	return worldObjects[x][y];
    }

    /**
     * Remove an object.
     */
    void removeWO(WorldObject wo) {
	worldObjects[wo.x][wo.y] = null;
    }

    /**
     * A <code>BasicBeast</code> explodes, due to being enclosed for a
     * certain period of time, taking out all 8 adjacent squares and itself.
     */
    void explode(BasicBeast beast) {
	int x = beast.x,
	    y = beast.y;

// 	System.err.println("Beast explodes! ("
// 			   + beast.x + ", " + beast.y + ")");

	for(int i = -1; i <= 1; i++) {
	    for(int j = -1; j <= 1; j++) {
		WorldObject wo = worldObjects[x + i][y + j];

		if(wo instanceof Beast) {
		    ((Beast)wo).dead = true;
		}

		// Heavy blocks are blast-proof.
		if(!(wo instanceof HeavyBlock)) {
		    worldObjects[x + i][y + j] = null;
		    clear(x + i, y + j);
		}
	    }
	}
    }

    void speedup() {
	BEAST_DELAY = (int)(speedup_factor * BEAST_DELAY);
	EGG_DELAY = (int)(speedup_factor * EGG_DELAY);
	EGG_HATCH_DELAY = (int)(speedup_factor * EGG_HATCH_DELAY);
// 	SPEEDUP_DELAY = (int)(speedup_factor * SPEEDUP_DELAY);

	pq.enqueue(new Speedup(this), SPEEDUP_DELAY);

	/* Maintain count == #beasts invariant. */
	count++;
    }

    /**
     * Increase the score when <code>Beast</code>s are killed.
     */
    void addScore(int score) {
	scores.add(score);
	updateStatus(false);
    }

    /**
     * Splash screen key handler.
     */
    private KeyListener splashKeyListener = new KeyAdapter() {
	    public void keyPressed(KeyEvent ke) {
// 		System.err.println("sKL.kP: " + ke);

		int keycode = ke.getKeyCode();

		if(KeyEvent.VK_A <= keycode && keycode <= KeyEvent.VK_Z) {
		    /* Letter selects a level. */
		    BeastsWorld.this.setLevel(keycode - KeyEvent.VK_A);
		} else switch(keycode) {
		case KeyEvent.VK_F1:
		    /* Display high scores. */
		    BeastsWorld.this.highScores();
		    break;
		case KeyEvent.VK_ESCAPE:
		    /* Exit game with escape. */
		    BeastsWorld.this.quit(QUIT_GAME);
		    break;
		case KeyEvent.VK_SPACE:
		    /* Start game with a space. */
		    synchronized(BeastsWorld.this) {
			BeastsWorld.this.notify();
		    }
		default:
		    /* Ignore all other keys. */
		    break;
		}
	    }
	};

    /**
     * Display splash screen.
     */
    void displaySplash() throws InterruptedException {
	update();

	/* Wait for the user to press a key. */
	this.addKeyListener(splashKeyListener);
	synchronized (this) {
	    this.wait();
	}
	this.removeKeyListener(splashKeyListener);
    }

    /**
     * Game loop. Display a splash screen, start the game itself, etc.
     * <p>
     * When the player runs out of lives, the thread that this code runs in
     * is interrupted.
     *
     * @see #gameOver
     */
    public void run() {
	thisThread = Thread.currentThread();

	/* Setup the off-screen graphics buffer. */
	if(offImage2 == null) {
	    offImage2 = createImage(dim.width, dim.height);
	    offGraphics = offImage2.getGraphics();
	}

	try {
	    /* Wait for the images to download. */
	    tracker.waitForAll();

	    /* Display the splash screen. */
	    offImage = splash;

	    while(quit != QUIT_GAME) {
		displaySplash();

		Levels.initLevel(this, level);
// 		count = 10;

		for(round = 1; round <= winningRound; round++) {
// 		    System.err.println("BW.run(): round: " + round
// 				       + " of " + winningRound);

		    player.initRound();
		    updateStatus(false);
		    boolean playerWins = playRound();
		    player.finishedRound();

		    /* Empty the PQ at this point to aid the GC. */
		    pq.empty();

		    if(!playerWins || quit > 0) {
			break;
		    } else {
			Levels.nextRound(this, level, round);
		    }
		}

		if(!training) {
		    scores.updateHighScores(Levels.levelToString(level, round));
		}

		switch(quit) {
		case 0:
		    scores.render(this);
		    break;

		case QUIT_LEVEL:
		    offImage = splash;
		    quit = 0;
		    break;

		case QUIT_GAME:
		    return;

		default:
		    throw new RuntimeException("unknown quit status: " + quit);
		}
	    }
	} catch (InterruptedException e) {
	    if(quit > 0) {
// 		System.err.println("interrupted exception quit");
		return;
	    }
	}

// 	System.err.println("run2 quit");
    }

    /**
     * @return false if the player is out of lives.
     */
    private boolean playRound() {
 	try {
	    while(count > 0 && quit == 0) {
		long sleep;

		if(thisThread.isInterrupted()) {
		    return false;
		}

		/* Sleep until the event should happen. */
		sleep = pq.getHighestPriority();

		if(sleep > 0) {
		    Thread.sleep(sleep);
		}
// 		Thread.sleep(0);

		Thread.yield();

		count--;

		/* Grab the lock on this object, to block the player. */
		synchronized(this) {
		    pq.dequeue().move();
		}

// 		System.err.println("BeastsWorld: count: " + count);
	    }
	} catch (PriorityQueueException e) {
	    /* Buggy use of the PQ. */
	    System.err.println(e);
	    throw new RuntimeException(e.toString());
	} catch (InterruptedException e) {
	    /* The player ran out of lives or they exited the game, and we
	     * were sleeping,  */
	    return false;
	}

	/* Player completed this round. */
	return true;
    }

    /**
     * Called by the <code>Player</code> when it runs out of lives.
     */
    void gameOver() {
// 	System.err.println("BeastsWorld.gameOver()");
	thisThread.interrupt();
    }

    /**
     * Asynchronous quit event.
     */
    synchronized void quit(int quitWhat) {
// 	System.err.println("BeastsWorld.quit()");
	quit = quitWhat;
	thisThread.interrupt();
    }

    /**
     * @arg keys Print out a second line of valid keys.
     */
    void updateStatus(boolean keys) {
	StringFormatter out = new StringFormatter();

	out.append("Level: ");
	out.append((char)(level + 'A') + String.valueOf(round),
		   3, StringFormatter.RIGHT, ' ');
	out.append(" Score: ");
	out.append(scores.getCurrentScore(),
		   4, StringFormatter.RIGHT, ' ');
	out.append(" Lives: ");
	out.append(String.valueOf(player.getLives()),
		   2, StringFormatter.RIGHT, ' ');

	drawString(out.toString(), 0, STATUSLINE);

	if(keys) {
	    drawString("SPACE = play / Letter = change levels / F1 = Scores / Esc = Exit", 6, STATUSLINE + 1);
	}

	update();
    }

    /**
     * Draw a string at the specified co-ordinates (treating the display
     * area as an 80x25 grid), using the font loaded into the field
     * <code>ascii</code> as an image strip (the first element has ASCII
     * code 33 "!").
     */
    void drawString(String string, int x, int y) {
// 	System.err.println("bW.drawString(): " + string + " x = " + x + " y = " + y);

	offGraphics.setColor(Color.white);

	/* Convert (x, y) into pixel co-ordinates. */
	x *= C_WIDTH;
	y *= C_HEIGHT;

	for(int i = 0; i < string.length(); i++) {
// 	    System.err.println("bW.drawString(): x' = " + x + " y' = " + y);
// 	    System.err.println("  x - str[i] * CWIDTH + 1 = " + (x - ((string.charAt(i) - 33) * C_WIDTH) + 1));

	    int c = string.charAt(i) - 33;

	    if(c < 0 || c > 125) {
		offGraphics.setColor(Color.black);
		offGraphics.fillRect(x, y, C_WIDTH, C_HEIGHT);
		offGraphics.setColor(Color.white);
	    } else {
		offGraphics.setClip(x, y, C_WIDTH, C_HEIGHT);
		offGraphics.drawImage(ascii,
				      x - (c * C_WIDTH) + 1, y,
				      null);
	    }
	    x += C_WIDTH;

	}

	offGraphics.setClip(0, 0, WIDTH, HEIGHT);
	update();
    }

    void highScores() {
	offImage = offImage2;
	scores.render(this);
	update();
    }

    public void paint(Graphics g) {
	update(g);
    }

    void update() {
	update(this.getGraphics());
    }

    public void update(Graphics g) {
// 	System.err.println("BeastsWorld.update()");
	g.drawImage(offImage, 0, 0, this);
    }

    /**
     * Package Java's random number generator.
     */
    int rnd(int limit) {
	return limit > 0 ? random.nextInt(limit) : 0;
    }

    double rnd() {
	return random.nextDouble();
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append(pq.toString());

	return out.toString();
    }
}
