package beasts;

/**
 * A priority-queue implementation. Assumes the keys are <code>long</code>s.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: PriorityQueue.java,v 1.5 2002/02/03 23:47:16 peteg Exp $
 */
class PriorityQueue {
    /**
     * The elements of a priority queue.
     */
    class PriorityQueueElement {
	/**
	 * The priority (lower numbers == higher priorities).
	 */
	long priority;

	/**
	 * The object at this priority level.
	 */
	Event o;

	PriorityQueueElement() {
	}

	void set(long priority, Event o) {
	    this.priority = priority;
	    this.o = o;
	}

	void clear() {
	    o = null;
	}

	public String toString() {
	    return "Priority: " + priority + " Element: " + o;
	}
    }

    /**
     * The heap, which we expand and contract as needs be.
     */
    private PriorityQueueElement[] heap;

    /**
     * The index of the last element of the heap.
     */
    private int last;

    PriorityQueue() {
	heap = new PriorityQueueElement[100];
	last = 0;
    }

    /**
     * Enqueue an object with a given priority.
     */
    void enqueue(Event o, int delay) {
	/* Adjust the priority by the current time. */
	long priority = delay + System.currentTimeMillis();

//  	System.err.println("PriorityQueue().enqueue: priority: " + priority + " - " + o);

	// Put the element at the end of the heap...
	if(++last >= heap.length) {
	    // Resize the array.
// 	    System.err.println("PriorityQueue.enqeue(): heap is full: resizing to: " + 2 * heap.length);

	    PriorityQueueElement[] old_heap = heap;
	    heap = new PriorityQueueElement[2 * old_heap.length];
	    System.arraycopy(old_heap, 0, heap, 0, old_heap.length);
	}

	PriorityQueueElement pqe;

	if((pqe = heap[last]) == null) {
// 	    System.err.println("Creating new PQE: last = " + last);
	    pqe = heap[last] = new PriorityQueueElement();
	}

	pqe.set(priority, o);

	// ...then bubble it up.
	for(int i = last; i > 1; i /= 2) {
	    if(heap[i/2].priority > priority) {
		// The new element has higher priority, swap.
		heap[i] = heap[i/2];
		heap[i/2] = pqe;
	    } else {
		// Finished: the heap has been re-established.
		break;
	    }
	}
    }

    /**
     * Dequeue the object with the highest priority, re-establish the heap.
     *
     * @exception PriorityQueueException thrown if the queue is empty.
     */
    Event dequeue() throws PriorityQueueException {
	if(isEmpty()) {
	    throw new PriorityQueueException("PriorityQueue.dequeue(): queue is empty.");
	}

// 	System.err.println("PriorityQueue.dequeue(): " + heap[1]);

	// The highest-priority element is at heap[1].
	Event o = heap[1].o;
	heap[1].clear();

	// Bubble down the right-most leaf from the root. Swap with the
	// removed element so that we don't leak memory.
	PriorityQueueElement pqe = heap[last];
	heap[last] = heap[1];
	last--;

	// Shrink the array if we need to (1/2 full -> shrink to 2/3 the size).
	if(heap.length > 100 && last < heap.length / 2) {
//  	    System.err.println("PriorityQueue.deqeue(): heap is less than half full: resizing to: " + heap.length * 2 / 3);

	    PriorityQueueElement[] old_heap = heap;
	    heap = new PriorityQueueElement[old_heap.length * 2 / 3];
	    System.arraycopy(old_heap, 0, heap, 0, heap.length);
	}

	int cur_index = 1, hpc_index = 2 * cur_index;

	while(hpc_index <= last) {
	    // Choose the highest priority child: Assume it's the left...
	    PriorityQueueElement hpc = heap[hpc_index];

	    if((hpc_index + 1 <= last) && heap[hpc_index + 1].priority < hpc.priority) {
		// ...no, it's the right.
		hpc_index++;
		hpc = heap[hpc_index];
	    }

	    // If it has higher priority than pqe...
	    if(hpc.priority < pqe.priority) {
		// ...move it up.
		heap[cur_index] = heap[hpc_index];

		// Next, consider the hpc node.
		cur_index = hpc_index;
		hpc_index *= 2;
	    } else {
		// ...otherwise we're done.
		break;
	    }
	}

	// Insert the bubble-down'ed element.
	heap[cur_index] = pqe;

	return o;
    }

    /**
     * Determine the priority of the highest-priority event in the
     * queue. This method is a hack to avoid <code>dequeue</code> having to
     * return some kind of tuple object.
     *
     * @exception PriorityQueueException thrown if the queue is empty.
     * @see #dequeue
     */
    int getHighestPriority() throws PriorityQueueException {
	if(isEmpty()) {
	    throw new PriorityQueueException("PriorityQueue.getHighestPriority(): queue is empty.");
	}

// 	System.err.println("PQ.gHP: " + heap[1]);

	/* Adjust for realtime. */
	return (int)(heap[1].priority - System.currentTimeMillis());
    }

    boolean isEmpty() {
	return last == 0;
    }

    void empty() {
	try {
	    while(!isEmpty()) {
		dequeue().signal();
	    }
	} catch (PriorityQueueException e) {
	    throw new RuntimeException(e.toString());
	}
    }

    /**
     * Signal all pending events.
     */
    void signal() {
	for(int i = 1; i <= last; i++) {
	    heap[i].o.signal();
	}
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	for(int i = 1; i <= last; i++) {
	    out.append("heap[");
	    out.append(i);
	    out.append("] = ");
	    out.append(heap[i]);
	    out.append("\n");
	}

	return out.toString();
    }
}
