package beasts;

/**
 * Basic beasts, employing machine learning to discover a strategy.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: RLBasicBeast.java,v 1.11 2002/03/20 08:16:24 peteg Exp $
 */
class RLBasicBeast extends BasicBeast {
    private static class RLBBState extends RLState {
	RLBBState(BeastsWorld beasts, RLearner rl, int x, int y) {
	    super(beasts, rl);
	}

	/**
	 * Compute the state number for the given world situation.
	 * <p>
	 * The lower eight bits are computed as follows: Given a location in
	 * the world and a vector, generate a bitstring that represents the
	 * immediately adjacent squares rotated such that the vector points
	 * "north" (i.e. == (0, -1)) in the transformed space.
	 * <code>Player</code> and <code>null</code> objects are mapped to
	 * 0, and everything else (all of which would block a
	 * <code>BasicBeast</code>) to 1.
	 * <p>
	 * The upper three bits are computed like so: Rotate the player
	 * vector by x/4 PI, where x is the int "direction" of the squashed
	 * player vector minus 1. (It's the number of times the grid needs
	 * to be rotated anti-clockwise to point the squashed vector
	 * "north".) From this, work out the distance to the player, and
	 * whether it's left, middle or right of "north".
	 *
	 * @arg x,y the beast's co-ordinates.
	 * @arg px,py the vector towards the player.
	 */
	int getKey(int x, int y,
		   int px, int py, int pdir) {
	    int dir;
	    int[] xy;
	    int tx, ty;
	    int key = 0;

	    /* Get the direction number for the square we want to be in the
	     * "northwestern" corner. (this is (-1) with wraparound - i.e. 0
	     * -> 7, 1 -> 0, etc.) */
	    dir = (pdir + 7) % 8;

	    /* Compute lower 8 bits. */

	    for(int i = 0; i < 8; i++) {
		xy = BeastsWorld.int2dir(dir); tx = xy[0]; ty = xy[1];

		WorldObject wo = this.beasts.getWO(x + tx, y + ty);

		if(!(wo == null || wo instanceof Player)) {
		    /* Obstacle: this bit should be 1. */
		    key |= 1 << i;
		}

		/* Rotate the vector clockwise for next direction. */
		dir = (dir + 1) % 8;
	    }

	    /* Compute upper 3 bits. */

	    /* Compute the taxi-metric distance to the player. */
	    int apx = BeastsWorld.abs(px),
		apy = BeastsWorld.abs(py);

	    int distance = apx > apy ? apx : apy;

	    distance = (distance <= 5)
		? 0 /* Near - within 5 squares. */
		: 1; /* Far - everywhere else. */

	    key |= distance << 8;

	    /* Compute the direction to the player. */

	    /* Use 2D rotation formula:
	     *   x' = x cos(theta) - y sin(theta)
	     *   y' = y cos(theta) + x sin(theta)
	     * with theta = -pdir/4 PI (clockwise rotation) */
	    double theta = - pdir * Math.PI / 4;
	    int pxr = (int)(px * Math.cos(theta) - py * Math.sin(theta) + 0.5),
		pyr = (int)(py * Math.cos(theta) + px * Math.sin(theta) + 0.5);

// 	System.err.println("gK: theta: " + theta
// 			   + " px, py = (" + px + ", " + py + ")"
// 			   + " pxr, pyr = (" + pxr + ", " + pyr + ")");

	    /* Use the rotated x co-ordinate to determine left, mid,
	     * right. */
	    dir = (pxr < 0) ? 0 /* Left. */
		: (pxr > 0) ? 2 /* Right. */
		: 1; /* Middle. */

	    key |= dir << 9;

// 		printKey(x, y, px, py, key);

	    return key;
	}

	private void printKey(int x, int y, int px, int py, int key) {
	    /* Debugging: print out the bits as a string, both original
	     * and transformed. */
	    System.err.println("Key: " + key
			       + " original vector (" + px + ", " + py + "): ");
	    for(int i = -1; i <= 1; i++) {
		for(int j = -1; j <= 1; j++) {
		    WorldObject wo = this.beasts.getWO(x + j, y + i);

		    System.err.print((wo == null || wo instanceof Player)
				     ? "." : "X");
		}
		System.err.println();
	    }
	    System.err.println("Transformed: ");
	    for(int i = 0; i < 8; i++) {
		if(i == 4) System.err.print(" ");
		System.err.print(((key & (1 << i)) > 0) ? "X" : ".");
	    }

	    int distance = key >> 8;
	    int dir = key >> 9;
	    System.err.println("Distance: " + (distance == 0 ? "Near" : "Far"));
	    System.err.println("Dir: " + ((dir == 0) ? "Left" : (dir == 1) ? "Right" : "Middle"));
	}

	int[] postProcessAction(int action, int pdir) {
// 	    System.err.println("RLBBrls: action: " + action);

	    /* Give reward if we moved towards the player. */
	    switch(action) {
	    case 0:
	    case 1:
	    case 2:
		giveReward(R_MOVED_TOWARDS_PLAYER);
		break;

	    case 3:
	    case 7:
// 		giveReward(R_MOVED_EAST_WEST);
// 		break;

	    case 4:
	    case 5:
	    case 6:
		giveReward(R_MOVED_AWAY_FROM_PLAYER);
		break;
	    }

	    /* Adjust move given the rotation performed by getKey(). */
// 	    action = (1 + pdir) % 8;
	    action = (action + pdir + 7) % 8;
// 	    System.err.println("  transformed action: " + action);

	    return BeastsWorld.int2dir(action);
	}
    }

    /**
     * Number of states:
     * <ul>
     *  <li> 8 adjacent squares (blocked, open)
     *  <li> 3 <i>x</i> directions (left, middle, right)
     *  <li> 2 <i>y</i> distances (near, far)
     * </ul>
     * so 2^8 * 3 * 2 = 1536
     */
    private final static int STATES = 1536;

    /**
     * Number of possible actions.
     */
    private final static int ACTIONS = 8;

//     private final static int INITIAL_Q_VAL = 100;
    private final static int INITIAL_Q_VAL = 0;

    /* Rewards. */
    private final static int R_INVALID_MOVE = -1000;
    private final static int R_DIED = -10;
    private final static int R_MOVED_TOWARDS_PLAYER = 1;
    private final static int R_MOVED_EAST_WEST = 1;
    private final static int R_MOVED_AWAY_FROM_PLAYER = -1;
    private final static int R_KILLED_PLAYER = 10;
    private final static int R_SAFE = 0; /* Safe move. */

    private RLState rls;

    RLBasicBeast(BeastsWorld beasts, int x, int y, RLearner rl) {
	super(beasts, x, y);
	this.rls = new RLBBState(beasts, rl, x, y);
    }

    void doMove(int dirs_count, int[] dirs, int px, int py) {
	int dx, dy;

	/* If we're adjacent to the player, simply move towards it. */
	int[] pxy = BeastsWorld.int2dir(BeastsWorld.squashdir2int(px, py));
	if(pxy[0] == px && pxy[1] == py) {
	    if(!vetMove(pxy[0], pxy[1])) {
		throw new RuntimeException("Dodgy stuff in RLBB.doMove()");
	    }
	} else {
	    /* Iterate until we make a valid move. */
	    while(true) {
		int[] xy = rls.getAction(x, y, px, py);
		dx = xy[0]; dy = xy[1];

		if(vetMove(dx, dy)) {
		    /* Valid move. */
// 		    System.err.println("RLBB: valid move: dx, dy = " + dx + ", " + dy);
		    break;
		} else {
		    /* Hit with immediate negative reward. */
// 		    System.err.println("RLBB: >> INVALID MOVE: dx, dy = " + dx + ", " + dy);
		    rls.giveReward(R_INVALID_MOVE);
		}
	    }
	}
    }

    void died() {
	super.died();
	rls.giveReward(R_DIED);
	rls.end();
    }

    void killed_player() {
	super.killed_player();
	rls.giveReward(R_KILLED_PLAYER);
// 	System.err.println("killed_player(): " + rls);
    }

    public void signal() {
// 	System.err.println("signal()");
	rls.end();
    }

    public String toString() {
	return "RLBasicBeast[" + x + ", " + y + "]";
    }

    static RLearner newRL() {
	return new RLearner(STATES, ACTIONS, INITIAL_Q_VAL);
    }
}
