package beasts;

/**
 * The (abstract) learner state kept on a
 * <code>Beast</code>-by-<code>Beast</code> basis.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: RLState.java,v 1.9 2002/03/18 07:23:18 peteg Exp $
 */
abstract class RLState {
    private class SAR {
	int state, action, reward;

	public String toString() {
	    StringBuffer out = new StringBuffer();

	    out.append("state: ");
	    out.append(state);
	    out.append(" action: ");
	    out.append(action);
	    out.append(" reward: ");
	    out.append(reward);

	    return out.toString();
	}
    }

    /**
     * Stack of <code>(state, action)</code> pairs.
     */
    private SAR[] actions;
    private int actions_top;
    private final static int ACTIONS_SIZE = 40;

    BeastsWorld beasts;
    private RLearner rl;

    RLState(BeastsWorld beasts, RLearner rl) {
	this.beasts = beasts;
	this.rl = rl;

	this.actions = new SAR[ACTIONS_SIZE];
	this.actions_top = 0;

	for(int i = 0; i < ACTIONS_SIZE; i++) {
	    actions[i] = new SAR();
	}
    }

    /**
     * Push a <code>(state, action, reward)</code> triple onto the stack.
     */
    void push(int state, int action, int reward) {
	SAR sar = actions[actions_top++];

	sar.state = state;
	sar.action = action;
	sar.reward = reward;

	if(actions_top == ACTIONS_SIZE) {
	    unwind_stack();
	}
    }

    private SAR pop() {
	return actions[--actions_top];
    }

    /**
     * Unwinds the <code>(state, action, reward)</code> stack.
     */
    private void unwind_stack() {
// 	System.err.println("unwind_stack(): " + this);
// 	Thread.dumpStack();

	if(actions_top > 0) {
	    SAR sar0 = pop(),
		sar_last = sar0;

	    while(actions_top > 0) {
		SAR sar = pop();
		rl.giveReward(sar.state, sar.action, sar.reward, sar0.state);
		sar0 = sar;
	    }

	    /* Put the last one back on the bottom of the stack. */
	    push(sar_last.state, sar_last.action, sar_last.reward);
	}
    }

    /**
     * Give a reward to the most-recent <code>(state, action)</code> pair.
     * Assumes rewards are additive, i.e. the final reward given to a
     * <code>(state, action)</code> pair is the sum of all rewards awarded
     * to it.
     */
    void giveReward(int reward) {
	if(actions_top == 0) {
	    /* No actions to reward! */
	    return;
	}

	actions[actions_top - 1].reward += reward;
    }

    /**
     * Given a world, decide what move (action) to take.
     *
     * @return a squashed vector, representing the direction to move in.
     */
    int[] getAction(int x, int y, int px, int py) {
	int pdir = BeastsWorld.squashdir2int(px, py);
	int key = getKey(x, y,
			 px, py, pdir);
	int action = rl.selectAction(beasts, key);

	push(key, action, 0);

	return postProcessAction(action, pdir);
    }

    /**
     * The learning task is over: push a "no successor state" indicator,
     * then empty the stack.
     */
    void end() {
// 	System.err.println("RLS.end()");
	push(-1, -1, 0);
	unwind_stack();
    }

    /**
     * Generate a new key.
     *
     * @arg x, y location of the beast.
     * @arg px, py player vector
     * @arg pdir the number of the squashed player vector.
     */
    abstract int getKey(int x, int y,
			int px, int py, int pdir);

    /**
     * Given an action in a possibly transformed domain, translate it into
     * an action in the "real" world.
     *
     * @arg pdir the number of the direction towards the player.
     * @return a squashed vector, representing the direction to move in.
     */
    abstract int[] postProcessAction(int action, int pdir);

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("RLState:\n");
	for(int i = 0; i < actions_top; i++) {
	    out.append("actions[");
	    out.append(i);
	    out.append("] = ");
	    out.append(actions[i]);
	    out.append("\n");
	}

	return out.toString();
    }
}
