package beasts;

import java.io.*;

/**
 * The reinforcement learning algorithm.
 * <p>
 * I use <i>Q</i>-learning as it is quite straightfoward to implement, and
 * the nature of the game satisfies the theoretical conditions for
 * convergence:
 * <ul>
 *  <li> Deterministic Markov process (well, probably not.)
 *  <li> Bounded reward function (easy.)
 *  <li> Each (state, action) pair is visited infinitely often (easy.)
 * </ul>
 * <p>
 * The update rule is as follows:<br>
 * <i>Q(s, a)' := Q(s, a) + alpha * (reward + gamma *</i> max<i><sub>a'</sub>[Q(s', a')] - Q(s, a))</i><br>
 * where:
 * <ul>
 *  <li> <i>s</i>: the current state.
 *  <li> <i>a</i>: the action taken, resulting in new state <i>s'</i>.
 *  <li> <i>alpha</i>: learning constant (0.1).
 *  <li> <i>gamma</i>: the discount on future rewards (0.9).
 *  <li> <i>reward</i>: the reward for making action <i>a</i> in state <i>s</i>, resulting in state <i>s'</i>.
 * </ul>
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: RLearner.java,v 1.10 2002/03/20 08:16:24 peteg Exp $
 */
class RLearner implements Serializable {
    static final long serialVersionUID = 1541947838151220168L;

    /* Action selection types. */
    private final static int E_GREEDY = 1;
    private final static int SOFTMAX = 2;

    private final static int ACTION_SELECTION = SOFTMAX;

    /* Chance of random selection using E_GREEDY. (1 chance in...) */
    private final static int EPSILON = 10;

    /* Learning step. */
    private final static double ALPHA = 0.1;

    /* Discount on future reward. */
    private final static double GAMMA = 0.9;

    /**
     * The <i>Q</i> values, which are what we're trying to learn.
     */
    private double[][] qvals;

    private int initial_q_val;

    RLearner(int numstates, int numactions, int initial_q_val) {
	this.qvals = new double[numstates][numactions];
	this.initial_q_val = initial_q_val;

	for(int s = 0; s < numstates; s++) {
	    for(int a = 0; a < numactions; a++) {
		setQ(s, a, initial_q_val);
	    }
	}
    }

    /* Assume vs.length > 0. */
    private static int max(double[] vs) {
	int index = 0;
	double max = vs[0];

	for(int i = vs.length; i-- > 0; ) {
	    if(vs[i] > max) {
		max = vs[i];
		index = i;
	    }
	}

	return index;
    }

    private double[] getQs(int key) {
	return qvals[key];
    }

    private void setQ(int key, int action, double qprime) {
	qvals[key][action] = qprime;
    }

    /**
     * Select an action using either the <i>e</i>-greedy or <i>softmax</i>
     * policy.
     */
    int selectAction(BeastsWorld beasts, int key) {
	double[] qvals = getQs(key);
 	int selectedAction = -1;

// 	System.err.println("selectMove: key: " + key);
// 	for(int i = 0; i < qvals.length; i++) {
// 	    System.err.println("qvals[" + i + "] = " + qvals[i]);
// 	}

	switch(ACTION_SELECTION) {
	case E_GREEDY:
	    if(beasts.rnd(EPSILON) == 0) {
		/* Explore. */
// 		System.err.println("sM: random.");
		selectedAction = beasts.rnd(qvals.length);
	    } else {
		/* Choose the one with the highest qval. If there are
		 * several, simply choose the first one in the array. */
// 		System.err.println("sM: max.");
		selectedAction = max(qvals);
	    }
	    break;
	
	case SOFTMAX:
	    /* p_j = exp(y_j) / (sum[i in 1..n] (exp(y_i))) */
	    double p[] = new double[qvals.length];
	    double denom = 0.0;

	    for(int i = 0; i < qvals.length; i++) {
		denom += p[i] = Math.exp(qvals[i]);
	    }

	    for(int i = 0; i < qvals.length; i++) {
		p[i] /= denom;
	    }

	    /* Select one based on the probabilities just computed. */
	    double rnd = beasts.rnd(),
		sum_p = 0.0;

	    for(int i = 0; i < qvals.length; i++) {
		if(sum_p < rnd && rnd < sum_p + p[i]) {
		    selectedAction = i;
		}
		sum_p += p[i];
	    }
	    break;

	default:
	    throw new RuntimeException("RLearner.selectMove(UNKNOWN)");
	}

// 	System.err.println("sM: selected action: " + selectedAction);
	return selectedAction;
    }

    /**
     * Given a world, a previous state of the world, an action and a reward,
     * update the qvals table.
     */
    void giveReward(int key, int action, int reward, int keyprime) {
// 	System.err.println("RL.gR(): key = " + key
// 			   + " action = " + action
// 			   + " reward = " + reward);

	double q = getQs(key)[action];
	double max_qprime_action;

// 	System.err.println("   q  = " + q);

	/* Get s', max a' */
	if(keyprime >= 0) {
	    double[] qsprime = getQs(keyprime);
	    max_qprime_action = qsprime[max(qsprime)];
	} else {
	    /* A negative keyprime signals no successor to this state.
	     * Hence max_qprime_action == 0.
	     */
	    max_qprime_action = 0;
	}

	setQ(key, action, q + ALPHA*(reward + GAMMA*max_qprime_action - q));

// 	System.err.println("   q' = " + getQs(key)[action]);
    }
}
