package beasts;

import java.awt.*;
import java.io.*;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * Track the current score, and the highest scores.
 * <p>
 * Serialization is used to make the high scores persistent, but this is
 * only possible when running as an application.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: Scores.java,v 1.6 2002/03/15 07:47:05 peteg Exp $
 */
class Scores implements Serializable {
    private static class Score implements Serializable {
	static final SimpleDateFormat df = new SimpleDateFormat("dd-MM-yyyy hh:mm");

	private String name = "< empty >";
	private String date = df.format(new Date());
	private String level = "";
	private int score = 0;

	Score() {
	}

	boolean lt(int score) {
	    return this.score < score;
	}

	void set(String name, String level, int score) {
	    this.name = name;
	    this.level = level;
	    this.score = score;

	    /* Date formatting is a little involved. */
	    this.date = df.format(new Date());
	}

	void render(BeastsWorld beasts, int i) {
	    StringFormatter out = new StringFormatter();

	    out.append(i, 2, StringFormatter.RIGHT, ' ');
	    out.append(": ");
	    out.append(name, 27, StringFormatter.LEFT, '.');
	    out.append(" ");
	    out.append(date);
	    out.append(" .. ");
	    out.append(score, 4, StringFormatter.RIGHT, ' ');
	    out.append(" .. ");
	    out.append(level);

	    beasts.drawString(out.toString(), 4, i + 1);
	}

	public String toString() {
	    return "Score: " + name + " date: " + date + " score: " + score + " level: " + level;
	}
    }

    private Score[] scores;
    transient int currentScore;

    private static final int SCORES = 20;

    Scores() {
	this.scores = new Score[SCORES];

	for(int i = 0; i < SCORES; i++) {
	    this.scores[i] = new Score();
	}
    }

    void reset() {
	currentScore = 0;
    }

    void add(int score) {
	currentScore += score;
    }

    int getCurrentScore() {
	return currentScore;
    }

    /**
     * Update high scores at the end of a level or on player death.
     */
    void updateHighScores(String level) {
	int i;

	for(i = SCORES - 1; i >= 0 && scores[i].lt(currentScore); i--);

	if(i < SCORES - 1) {
	    /* Shuffle all the lower scores down. */
	    for(int j = SCORES - 1; i + 1 < j; j--) {
		scores[j] = scores[j - 1];
	    }

	    Score score = scores[i+1] = new Score();
	    String name;

	    try {
		name = System.getProperty("user.name");
	    } catch (SecurityException e) {
		name = "< unknown applet user >";
	    }

	    score.set(name,
		      level,
		      currentScore);
	}
    }

    /**
     * Render the high score table on the canvas.
     */
    void render(BeastsWorld beasts) {
	beasts.clear();

	/* Add a border of immovable block images. */
	for(int i = 0; i < BeastsWorld.COLS; i++) {
	    beasts.draw(i, 0,                    HeavyBlock.image);
	    beasts.draw(i, BeastsWorld.ROWS - 1, HeavyBlock.image);
	}

	for(int i = 1; i < BeastsWorld.ROWS - 1; i++) {
	    beasts.draw(0,                    i, HeavyBlock.image);
	    beasts.draw(BeastsWorld.COLS - 1, i, HeavyBlock.image);
	}

	/* Add the header. */
	beasts.drawString("Name                Date     Time   Score    Level", 19, 1);

	/* Add the scores. */
	for(int i = 0; i < SCORES; i++) {
	    scores[i].render(beasts, i + 1);
	}

	beasts.updateStatus(true);
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("Current score: ");
	out.append(currentScore);
	out.append("\nSCORES\n======\n");
	for(int i = 0; i < SCORES; i++) {
	    out.append(scores[i].toString());
	    out.append("\n");
	}

	return out.toString();
    }
}
