/*
 * SETGID wrapper for BEASTS.
 *
 * Install this SETGID (group owned by "games"). It fires off the Java
 * interpreter with these permissions, after sanitizing the environment.
 *
 * Under Linux (at least), the following will tell you what your programs
 * are running as:
 *
 * ps -o "user,euser,group,egroup,cmd"
 *
 * $Id: beasts.c,v 1.3 2002/03/20 08:16:24 peteg Exp $
 */
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static char *java = "/usr/bin/java";
static char *class = "beasts.Beasts";
static char *jar = "CLASSPATH=/usr/share/java/beasts.jar";
static char *high_scores = "/var/games/beasts/beasts_state.ser";

void print_perms(char *name)
{
    fprintf(stderr, "%s: pid: %d user: %d group: %d\n",
	    name, getpid(), geteuid(), getegid());
}

char *findevar(char *envp[], char *key)
{
    int i, e;
    char *elt;

    for(e = 0; (elt = envp[e]) && elt != NULL; e++) {
	for(i = 0; key[i] != '\0' && key[i] == elt[i]; i++);
	if(key[i] == '\0' && elt[i] == '=')
	    break;
    }

    return elt;
}

/*
 * Execute the program, dropping the setgid privilege.
 */
int exec(int argc, char *argv[], char *envp[])
{
    int status;

    char *argvc[4];
    char *envpc[6];

    argvc[0] = java;
    argvc[1] = class;

    if(argc > 1 && !strcmp(argv[1], "-t")) {
	argvc[2] = "-t";
	argvc[3] = NULL;
    } else {
	argvc[2] = NULL;
    }

    envpc[0] = jar;
    envpc[1] = findevar(envp, "DISPLAY");
    envpc[2] = findevar(envp, "HOME");
    envpc[3] = "PATH=/bin:/usr/bin";
    envpc[4] = findevar(envp, "USER");
    envpc[5] = NULL;

    /* Drop setgid privilege. */
    setegid(getgid());

/*     print_perms("child"); */

    status = execve(java, argvc, envpc);
    fprintf(stderr, "%s: trying to execute %s: %s\n",
	    argv[0], java, strerror(errno));
    return status;
}

int main(int argc, char *argv[], char *envp[])
{
    pid_t pid;
    int retval;
    int stdinfds[2], stdoutfds[2];

    if(pipe(stdinfds) == -1 || pipe(stdoutfds) == -1) {
	fprintf(stderr, "%s: create pipe: %s",
		argv[0], strerror(errno));
	exit(1);
    }

    if((pid = fork()) == 0) {
	/* Child. */

	/* Hookup stdin, stdout to the pipes. */
	close(stdinfds[1]);
	close(stdoutfds[0]);

	dup2(stdinfds[0], 0);
	dup2(stdoutfds[1], 1);

	retval = exec(argc, argv, envp);
    } else {
	/* Parent. */
	int buf[BUFSIZ];
	int fd;
	int count;

	/* Check for fork failure. */
	if(pid == -1) {
	    fprintf(stderr, "%s: fork failure: %s",
		    argv[0], strerror(errno));
	    exit(1);
	}

/* 	print_perms("parent"); */

	close(stdinfds[0]);
	close(stdoutfds[1]);

	/* Hose the existing umask so we get what we ask for. */
	umask(0);

	/* Send the highscore table down the pipe. */
	if((fd = open(high_scores, O_RDONLY)) == -1) {
	    fprintf(stderr, "%s: trying to open-for-read/create %s: %s",
		    argv[0], high_scores, strerror(errno));
	    exit(1);
	}

	while((count = read(fd, buf, BUFSIZ)) > 0) {
	    if(write(stdinfds[1], buf, count) == -1) {
		fprintf(stderr, "%s: write to pipe: %s",
			argv[0], strerror(errno));
		exit(1);
	    }
	}

	if(count == -1) {
	    fprintf(stderr, "%s: read from file: %s",
		    argv[0], strerror(errno));
	    exit(1);
	}

	close(fd);
	close(stdinfds[1]);

	/* Block, waiting for the highscore table to be returned to us.
	 * To do things properly, we should:
	 *   1. Lock the file (fcntl(2)).
	 *   2. Merge high scores, rather than simply overwrite the whole
	 *      table.
	 */
	if((count = read(stdoutfds[0], buf, BUFSIZ)) == -1) {
	    fprintf(stderr, "%s: trying to read pipe: %s",
		    argv[0], strerror(errno));
	    exit(1);
	}

	if((fd = open(high_scores, O_WRONLY)) == -1) {
	    fprintf(stderr, "%s: trying to open-for-write/create %s: %s",
		    argv[0], high_scores, strerror(errno));
	    exit(1);
	}

	do {
	    if(write(fd, buf, count) == -1) {
		fprintf(stderr, "%s: writing to file %s: %s",
			argv[0], high_scores, strerror(errno));
		exit(1);
	    }
	} while((count = read(stdoutfds[0], buf, BUFSIZ)) > 0);

	if(count == -1) {
	    fprintf(stderr, "%s: read from file: %s",
		    argv[0], strerror(errno));
	    exit(1);
	}

	close(fd);
	close(stdoutfds[0]);
    }

    return retval;
}
